﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.Linq;
using System.Web;
using Dynamicweb.Extensibility;
using Dynamicweb.Extensibility.Provider;
using Dynamicweb.Ecommerce.Prices;
using Dynamicweb.Ecommerce.Products;
using Dynamicweb.Modules.UserManagement;
using Dynamicweb.Ecommerce.Integration;
using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Security.UserManagement;
using Dynamicweb.Core;

namespace Dynamicweb.Ecommerce.LiveIntegration
{
	public class PrepareProductInfoProvider : Dynamicweb.Ecommerce.Prices.PriceProvider
	{
		private static CacheLevel ProductCacheLevel
		{
			get
			{
				string cacheLevelString = Helpers.GetConfigValue(Constants.CacheConfiguration.Path, Constants.CacheConfiguration.Product);
				return Helpers.GetEnumValueFromString<CacheLevel>(cacheLevelString, CacheLevel.Page);
			}
		}

		private static bool AddProductFieldsToRequest
		{
			get
			{
				return Convert.ToBoolean(Helpers.GetConfigValue(Constants.CustomFieldsSettings.Path, Constants.CustomFieldsSettings.AddProductFieldsToRequest));
			}
		}

		public override Prices.PriceRaw FindPrice(Product Product, double Quantity, string VariantID, International.Currency Currency, string UnitID, User User)
		{
			if (!Configuration.Global.IsIntegrationActive)
				return null;

			if (Convert.ToBoolean(Helpers.GetConfigValue(Constants.WebServiceConfiguration.Path, Constants.WebServiceConfiguration.DisableLiveIntegrationOnProductCatalog)))
				return null;

			if (!Configuration.Global.IsIntegrationForAnonymousUsersAllowed && User == null)
				return null;

			if (User != null && User.IsLivePricesDisabled)
				return null;

			if (Product == null || Product.Id == "" || Product.Number == "")
				return null;            

			try
			{                                
                // If Product is not in the cache, then get it from Erp
                if (!ErpResponseCache.IsProductInCache(ProductCacheLevel, Helpers.ProductIdentifier(Product)))
				{
					List<Product> products = new List<Product>();
                    products.Add(Product);                    
                    if (FetchProductInfos(products))
					{                                                
                        //Check if requested product was not received from response
                        if (!ErpResponseCache.IsProductInCache(ProductCacheLevel, Helpers.ProductIdentifier(Product)))
						{
							Logger.Instance.Log(ErrorLevel.Error, string.Format("Error receiving product info for product: {0}.", Helpers.ProductIdentifier(Product)));
							return null;
						}
					}
					else
					{                     
						return null;
					}
				}

				ProductInfo priceInfo = ErpResponseCache.GetProductInfos(ProductCacheLevel)[Helpers.ProductIdentifier(Product)];

				// Init PriceRaw
				var priceRaw = new PriceRaw();
				priceRaw.Currency = Currency;

				// Set PriceRaw price
				priceRaw.Price = (double)priceInfo["TotalPrice"];

				return priceRaw;
			}
			catch (Exception e)
			{
				Logger.Instance.Log(ErrorLevel.Error, string.Format("Unknown error during FindPrice(). Exception: {0}", e.Message));
				return null;
			}            
        }

		public override void PreparePrices(Dictionary<Product, double> Products)
		{
			// This overload is invoked when OrderLines are prepared (only).

			return;
		}

		public override void PreparePrices(ProductCollection Products)
		{
			// This overload is invoked when products are rendered.            

			if (Configuration.Global.IsIntegrationActive && Products != null
						   && (string.IsNullOrEmpty(Helpers.GetConfigValue(Constants.WebServiceConfiguration.Path, Constants.WebServiceConfiguration.Shop)) ||
	Helpers.GetConfigValue(Constants.WebServiceConfiguration.Path, Constants.WebServiceConfiguration.Shop) == (String)Dynamicweb.Frontend.PageView.Current().Area.EcomShopId))
			{
				FetchProductInfos(Products.ToList());
			}
		}

		public static bool FetchProductInfos(List<Product> products)
		{
			if (products == null || products.Count == 0)
				return false;

			if (Convert.ToBoolean(Helpers.GetConfigValue(Constants.WebServiceConfiguration.Path, Constants.WebServiceConfiguration.DisableLiveIntegrationOnProductCatalog)))
			{
				Logger.Instance.Log(ErrorLevel.DebugInfo, "Live integration is disabled for Product Catalog.");
				return false;
			}
			var user = User.GetCurrentExtranetUser();
			if (!Global.IsIntegrationForAnonymousUsersAllowed && user == null)
			{
				Logger.Instance.Log(ErrorLevel.DebugInfo, "No user is currently logged in. Anonymous user is not allowed.");
				return false;
			}

			if (user != null && user.IsLivePricesDisabled)
			{
				Logger.Instance.Log(ErrorLevel.DebugInfo, string.Format("Calculated prices are not allowed to the user '{0}'.", User.GetCurrentExtranetUser().UserName));
				return false;
			}

			// Check for existance of the given products in the Cache
			List<Product> productsForRequest = new List<Product>();            
			foreach (Product product in products)
			{                             
                if (!ErpResponseCache.IsProductInCache(ProductCacheLevel, Helpers.ProductIdentifier(product)))
					productsForRequest.Add(product);
			}

			if (productsForRequest.Count > 0)
			{
				XmlDocument request = BuildProductRequest(productsForRequest);                

				XmlDocument response = Connector.GetProductsInfo(request.OuterXml);

				if (response != null && !string.IsNullOrEmpty(response.InnerXml))
				{
					// Parse the response
					Dictionary<string, ProductInfo> prices = ProcessResponse(response);

					if (prices != null && prices.Count > 0)
					{
						// Cache prices
						foreach (string productKey in prices.Keys)
						{
							Dictionary<string, ProductInfo> piCache = ErpResponseCache.GetProductInfos(ProductCacheLevel);
							if (piCache.ContainsKey(productKey))
								piCache.Remove(productKey);
							piCache.Add(productKey, prices[productKey]);
						}
					}
					else
					{
						return false;
					}
				}
				else
				{
					//error occured
					return false;
				}
			}
			return true;
		}

		// Fills the values from Response on the Product
		public static void FillProductValues(Product product)
		{
			if (product != null && ErpResponseCache.IsProductInCache(ProductCacheLevel, Helpers.ProductIdentifier(product)))
			{
				Dictionary<string, ProductInfo> piCache = ErpResponseCache.GetProductInfos(ProductCacheLevel);
				if (piCache.ContainsKey(Helpers.ProductIdentifier(product)))
				{
					ProductInfo productInfo = piCache[Helpers.ProductIdentifier(product)];
					product.Stock = (double)productInfo["Stock"];
					product.Price.PriceWithVAT = (double)productInfo["TotalPrice"];
					//Update Product Custom Fields
					if (AddProductFieldsToRequest && product.ProductFieldValues.Count > 0)
					{
						foreach (ProductFieldValue pfv in product.ProductFieldValues)
						{
							if (productInfo.ContainsKey(pfv.ProductField.SystemName))
								pfv.Value = productInfo[pfv.ProductField.SystemName];
						}
					}
				}
			}
		}

		#region Erp build request and parse response

		private static XmlDocument BuildProductRequest(List<Product> products)
		{
			if (products == null || products.Count == 0)
				return null;

			User user = User.GetCurrentExtranetUser();
			string externalUserID = Helpers.CustomerIdentifier;

			string currencyCode = Ecommerce.Common.Context.Currency.Code;

			// Create XmlDocument object with Declaration
			XmlDocument xDoc = new XmlDocument();
			XmlDeclaration xDec = xDoc.CreateXmlDeclaration("1.0", "UTF-8", "yes");
			xDoc.AppendChild(xDec);
			// Create Request root node
			XmlElement xRoot = xDoc.CreateElement("GetEcomData");
			xRoot.SetAttribute("ExternalUserID", externalUserID);

			XmlElement tablesNode = xDoc.CreateElement("tables");
			// Add Products
			XmlElement productsNode = xDoc.CreateElement("Products");
			productsNode.SetAttribute("type", "filter");

            var doCalculatePriceUsingProductNumber = Converter.ToBoolean(Helpers.GetConfigValue(Constants.PriceCalculation.Path, Constants.PriceCalculation.CalculatePriceUsingProductNumber));

            foreach (Product product in products)
			{
				if (product != null)
				{
					XmlElement productNode = xDoc.CreateElement("Product");

                    string productNumber = Helpers.GetProductNumber(product);
					if (doCalculatePriceUsingProductNumber)
					{
                        Helpers.AddChildXmlNodeToProductList(xDoc, productNode, "ProductID", productNumber);
                    }
					else
					{
						Helpers.AddChildXmlNodeToProductList(xDoc, productNode, "ProductID", product.Id);                        
                    }

                    Helpers.AddChildXmlNodeToProductList(xDoc, productNode, "ProductNumber", productNumber);
					Helpers.AddChildXmlNodeToProductList(xDoc, productNode, "CurrencyCode", currencyCode);

					if (AddProductFieldsToRequest && product.ProductFieldValues.Count > 0)
					{
						XmlElement fieldsNode = xDoc.CreateElement("Fields");
						foreach (ProductFieldValue pfv in product.ProductFieldValues)
						{
							Helpers.AddChildXmlNodeToProductList(xDoc, fieldsNode, pfv.ProductField.SystemName, pfv.Value != null ? pfv.Value.ToString() : string.Empty);
						}
						productNode.AppendChild(fieldsNode);
					}

					// Add nodes to the respective parent nodes    
					productsNode.AppendChild(productNode);
				}
			}

			tablesNode.AppendChild(productsNode);
			xRoot.AppendChild(tablesNode);
			xDoc.AppendChild(xRoot);

			return xDoc;
		}

		// Processes the response
		private static Dictionary<string, ProductInfo> ProcessResponse(XmlDocument response)
		{
			if (response == null)
				return null;

			try
			{
				Dictionary<string, ProductInfo> infos = new Dictionary<string, ProductInfo>();

				foreach (XmlNode item in response.SelectNodes("//item [@table='EcomProducts']"))
				{
					ProductInfo productInfo = new ProductInfo();
					// Set ProductID
					productInfo["ProductID"] = Helpers.ProductIdentifier(item.SelectSingleNode("column [@columnName='ProductID']").InnerText,
						Common.Context.LanguageID);
					// Set price
					productInfo["TotalPrice"] = Helpers.ToDouble(item.SelectSingleNode("column [@columnName='ProductPrice']").InnerText);
					// Set stock
					productInfo["Stock"] = Helpers.ToDouble(item.SelectSingleNode("column [@columnName='ProductStock']").InnerText);
					//Set ProductCustomFields
					if (AddProductFieldsToRequest)
					{
						XmlNode fieldNode = null;
						foreach (ProductField pf in ProductField.GetProductFields())
						{
							fieldNode = item.SelectSingleNode(string.Format("column [@columnName='{0}']", pf.SystemName));
							if (fieldNode != null)
							{
								productInfo[pf.SystemName] = fieldNode.InnerText;
							}
						}
					}
					infos.Add((string)productInfo["ProductID"], productInfo);
				}

				return infos;
			}
			catch (Exception e)
			{
				Logger.Instance.Log(ErrorLevel.Error, string.Format("Response does not match schema: '{0}'.", e.Message));
			}

			return null;
		}

		#endregion
	}
}
